import inspect
import itertools
import re
from typing import overload

import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.axes import Axes

from .validation import validate

support_seaborn = []
for name in dir(sns):
    func = getattr(sns, name)
    if callable(func) and "ax" in inspect.signature(func).parameters:
        support_seaborn.append(name)


class LazyAxes(Axes):

    attrs = ["attr", "next", "args", "kwargs"]
    methods = [
        "reverse", "sns", "seaborn", "_ipython_canary_method_should_not_exist_"
    ]

    def __init__(self, attr=None):
        self.attr = attr
        self.next = None
        self.args = []
        self.kwargs = {}

    def __getattribute__(self, attr):
        if attr in ["attrs", "methods"]:
            return super().__getattribute__(attr)
        if attr in self.attrs + self.methods:
            return super().__getattribute__(attr)

        if self.attr is None:
            self.attr = attr
            return self
        else:
            self.next = LazyAxes(attr)
            return self.next

    def __call__(self, *args, **kwargs):
        self.is_called = True
        self.args = args
        self.kwargs = kwargs
        return self

    def __str__(self):
        return f"<LazyAxes: {self.attr}>"

    def __repr__(self):
        return f"<LazyAxes: {self.attr}>"

    def __iter__(self):
        yield self

    def reverse(self, ax):
        lazy_ax = self
        while lazy_ax:
            if lazy_ax.attr == "seaborn":
                lazy_ax = lazy_ax.next
                func = lazy_ax.attr
                if func not in support_seaborn:
                    raise ValueError(
                        f"{func} is not suppoted. choose form {support_seaborn}")

                lazy_ax.kwargs["ax"] = ax
                ax = getattr(sns, func)(*lazy_ax.args, **lazy_ax.kwargs)
            else:
                ax = getattr(ax, lazy_ax.attr)
                if lazy_ax.args or lazy_ax.kwargs:
                    ax = ax(*lazy_ax.args, **lazy_ax.kwargs)
            lazy_ax = lazy_ax.next
        return ax

    @property
    def seaborn(self) -> sns:
        if sns is None:
            raise ModuleNotFoundError("No module named 'seaborn'.")

        self.attr = "seaborn"
        return self

    @property
    def sns(self, *args, **kwargs) -> sns:
        return self.seaborn(*args, **kwargs)


class figure:

    def __init__(self, level: str = "warning", **kwargs):
        self.level = level
        self.kwargs = kwargs

        self.keys: list[str] = []
        self.lazyaxes: list[LazyAxes] = []
        self.lazyattrs: list[LazyAxes] = []

    def __getattr__(self, name):
        fig_attr = LazyAxes(name)
        self.lazyattrs.append(fig_attr)
        return fig_attr

    def __getitem__(self, label: str) -> LazyAxes:
        if not isinstance(label, str):
            raise KeyError("support str only.")
        ax = LazyAxes()
        self.keys.append(label)
        self.lazyaxes.append(ax)
        return ax

    def draw(self, mosaic):
        mosaic = convert_mosaic(mosaic)
        chain = set(itertools.chain(*mosaic))

        fig = plt.figure(**self.kwargs)
        ax_dict = fig.subplot_mosaic(mosaic)

        for key, lazy_ax in zip(self.keys, self.lazyaxes):
            for match_key in filter(re.compile(key).search, chain):
                lazy_ax.reverse(ax_dict[match_key])

        for fig_attr in self.lazyattrs:
            fig_attr.reverse(fig)

        validate(fig, level=self.level)
        return fig

    def show(self, mosaic):
        self.draw(mosaic)
        # TODO: これをオブジェクト指向的にやる方法って無いのかな
        plt.show()

    def save(self, filename):
        self.draw()
        plt.savefig(filename)

    def __str__(self):
        return f"<figure: {id(self)}>"


def convert_mosaic(mosaic):
    if isinstance(mosaic, str) and "|" in mosaic:
        return [[cell.strip() for cell in line.split("|")]
                for line in mosaic.splitlines()]
    return mosaic



import warnings
from pathlib import Path

import numpy as np
from matplotlib.testing.compare import calculate_rms
from oekaki.validation import MisleadingWarning
from PIL import Image


def compare_figures(test_func):

    def wrapper():
        warnings.simplefilter("ignore", MisleadingWarning)

        expected_img, actual_img = map(convert_to_ndarray, test_func())

        res = calculate_rms(expected_img, actual_img)
        if res > 0:
            save_diff_image(expected_img, actual_img,
                            f"tests/failed_cases/{test_func.__name__}.png")
            raise AssertionError("over RMES.")

    return wrapper


def convert_to_ndarray(fig):
    fig.canvas.draw()
    data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    w, h = fig.canvas.get_width_height()
    return data.reshape((h, w, 3))


def save_diff_image(expected_image, actual_image, output):
    expected_image = np.array(expected_image).astype(float)
    actual_image = np.array(actual_image).astype(float)
    abs_diff_image = np.abs(expected_image - actual_image)

    # expand differences in luminance domain
    abs_diff_image *= 255 * 10
    save_image_np = np.clip(abs_diff_image, 0, 255).astype(np.uint8)
    height, width, depth = save_image_np.shape

    # The PDF renderer doesn't produce an alpha channel, but the
    # matplotlib PNG writer requires one, so expand the array
    if depth == 3:
        with_alpha = np.empty((height, width, 4), dtype=np.uint8)
        with_alpha[:, :, 0:3] = save_image_np
        save_image_np = with_alpha

    # Hard-code the alpha channel to fully solid
    save_image_np[:, :, 3] = 255

    path = Path(output).parent
    if not path.exists():
        path.mkdir(parents=True)
    Image.fromarray(save_image_np).save(output, format="png")



import warnings
from pathlib import Path

import matplotlib
import numpy as np
import oekaki
import seaborn as sns
from oekaki.validation import MisleadingWarning
from pytest import PytestUnknownMarkWarning

from tests.utils import compare_figures

matplotlib.use('Agg')
import matplotlib.pyplot as plt  # noqa

warnings.simplefilter("ignore", (MisleadingWarning, PytestUnknownMarkWarning))

path = Path(__file__).parent.joinpath("failed_cases")
[p.unlink() for p in path.glob("*.png") if p.is_file()]

x = np.linspace(0, 2 * np.pi, 50)


@compare_figures
def test_single():
    mosaic = "A"
    expected = plt.figure()
    ax_dict = expected.subplot_mosaic(mosaic)
    ax_dict["A"].bar(["a", "b", "c"], [5, 7, 9])

    actual = oekaki.figure(level="ignore")
    actual["A"].bar(["a", "b", "c"], [5, 7, 9])
    actual = actual._draw(mosaic)
    return expected, actual


@compare_figures
def test_line():
    mosaic = "AB"
    expected = plt.figure()
    ax_dict = expected.subplot_mosaic(mosaic)
    ax_dict["A"].bar(["a", "b", "c"], [5, 7, 9])
    ax_dict["B"].plot([1, 2, 3])

    actual = oekaki.figure(level="ignore")
    actual["A"].bar(["a", "b", "c"], [5, 7, 9])
    actual["B"].plot([1, 2, 3])
    actual = actual._draw(mosaic)
    return expected, actual


@compare_figures
def test_table():
    mosaic = "AB\nCD"
    rand = np.random.rand(100)

    expected = plt.figure()
    ax_dict = expected.subplot_mosaic(mosaic)
    ax_dict["A"].bar(["a", "b", "c"], [5, 7, 9])
    ax_dict["B"].plot([1, 2, 3])
    ax_dict["C"].hist(rand)
    ax_dict["D"].imshow([[1, 2], [2, 1]])

    actual = oekaki.figure(level="ignore")
    actual["A"].bar(["a", "b", "c"], [5, 7, 9])
    actual["B"].plot([1, 2, 3])
    actual["C"].hist(rand)
    actual["D"].imshow([[1, 2], [2, 1]])
    actual = actual._draw(mosaic)
    return expected, actual


@compare_figures
def test_sns():
    mosaic = "AB"
    expected = plt.figure()
    ax_dict = expected.subplot_mosaic(mosaic)
    ax_dict["A"].bar(["a", "b", "c"], [5, 7, 9])
    sns.lineplot(data=[1, 2, 3], ax=ax_dict["B"])

    actual = oekaki.figure(level="ignore")
    actual["A"].bar(["a", "b", "c"], [5, 7, 9])
    actual["B"].sns.lineplot(data=[1, 2, 3])
    actual = actual._draw(mosaic)
    return expected, actual


@compare_figures
def _test_table_all_slice():
    expected, axes = plt.subplots(2, 2, squeeze=False)
    for ax in axes.reshape(-1):
        ax.set(xlabel="あ", ylabel="い")
    axes[1, 1].plot(x, np.sin(x))

    actual = oekaki.figure(level="ignore")
    actual[:].set(xlabel="あ", ylabel="い")
    actual[1, 1].plot(x, np.sin(x))
    actual, axes = actual._draw()
    return expected, actual


# TODO: test validations



import re
import warnings
from typing import Literal

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

import oekaki

UNIT_BRACKET = "(", ")"


def show(level: str = "error"):
    validate(plt.gcf(), level)
    plt.show()


class MisleadingWarning(DeprecationWarning):
    pass


class validate:

    def __init__(self, fig: Figure, level: Literal["error", "warning", "ignore"]):
        if level not in {"error", "warning", "ignore"}:
            raise ValueError(
                'invalid level. choose from "error", "warning", or "ignore".')

        self.fig = fig
        self.warnings = []

        level = level.replace("warning", "always")
        warnings.simplefilter(level, MisleadingWarning)
        self.validate()

    def validate(self):
        for ax in self.fig.get_axes():
            if not ax.has_data():
                continue
            self.validate_labels(ax)
        if self.warnings:
            warnings.warn(
                "".join(f'\n{w["label"]}: {w["msg"]}' for w in self.warnings),
                MisleadingWarning, stacklevel=5)

    def validate_labels(self, ax: Axes):
        _1, _2 = oekaki.UNIT_BRACKET
        label = ax.get_label()

        for var in ["x", "y"]:
            target = getattr(ax, f"get_{var}label")()
            if not target:
                self.warnings.append({"label": label, "msg": f"No {var}-label."})
            elif not re.search(f".*\\{_1}.*\\{_2}", target):
                # TODO: English
                self.warnings.append({"label": label, "msg": f"No {var}-unit."})

            getattr(ax, f"set_{var}label")(target.replace(f"{_1}{_2}", ""))
